--[[
********************************
Licence: GPL
(c) 2012 msl
thx rudi for cordic & script check
v0.4
********************************
@title Twilight

@param a aktuelles Datum?
@default a 1
@range a 0 1

@param y Jahr
@default y 0
@values y 2012 2013 2014 2015 2016 2017 2018 2019 2020

@param m Monat
@default m 0
@values m Jan Feb Mar Apr Mai Jun Jul Aug Sep Okt Nov Dez

@param d Tag
@default d 30
@range d 1 31

@param s Sommerzeit?
@default s 0
@range s 0 1
]]

--FUNCTIONS------------------------------

-- Split string and write in array  -> string.split(string,delimiter)
function string:split(delimiter)
    local result = { }
    local from = 1
    local delim_from, delim_to = string.find(self, delimiter, from )
    while delim_from do
        table.insert( result, string.sub(self, from , delim_from-1 ) )
        from = delim_to + 1
        delim_from, delim_to = string.find(self, delimiter, from )
    end
    table.insert( result, string.sub(self, from ) )
    return result
end

-- remove leading & trailing spaces -> string.trim(string)
function string:trim()
  return (string.gsub(self, "^%s*(.-)%s*$", "%1"))
end

-- a x b
function mul_1E3(a, b)
    return cordic.integer_mul_div(a, b, SCALE)
end

-- a / b
function div_1E3(a, b)
    return cordic.integer_mul_div(a, SCALE, b)
end

function load_data(dfile)
    if os.stat(dfile) then
        local file = io.open(dfile)
        local id = table.getn(location)
        local line_id = id
                for line in file:lines(dfile) do
                    if string.find(line, "#") == nil then
                        local array = string.split(line,";")
                        if table.getn(array) > 3 then
                            array[2] = tonumber(array[2])
                            array[3] = tonumber(array[3])
                            array[4] = tonumber(array[4])
                            if (type(array[1]) == "string") then
                                array[1] = string.trim(array[1])
                                if (type(array[2]) == "number") and (array[2] <= 1800 or array[2] >= -1800) then
                                    if (type(array[3]) == "number") and (array[3] <= 1800 or array[3] >= -1800) then
                                        if (type(array[4]) == "number") and (array[4] <= 12   or array[4] >= -12)   then
                                            line_id = line_id + 1
                                            location[line_id] = {}
                                            location[line_id].name = array[1]
                                            location[line_id].lat  = array[2]
                                            location[line_id].lng  = array[3]
                                            location[line_id].utc  = array[4]
                                        end
                                    end
                                end
                            end
                        end
                    end
                end
        file:close()
        return true, line_id - id
    else
        return false
    end
end

function call_data()
    cls()
    print("    Datei-Browser wird geffnet.")
    print("       Whle eine Datei aus.")
    console_redraw()
    sleep(1500)
    cls()
    data_file = file_browser("A/CHDK/DATA")
    if data_file ~= nil then
        data, count = load_data(data_file)
    else
        data = false
    end
    if data == false then
        print("Konnte keine externen Daten laden!")
    else
        print("    ", count, "Datenstze eingelesen.")
    end
    console_redraw()
    sleep(1000)
end

function menu_geo_data()
    while true do
        cls()
        print(" [SET]    Lade externe Geo-Daten")
        print(" [\26]      Eingabe Geo-Daten")
        print(" [\27]      Lsche Datensatz")
        print(" [DISP]   GPS")
        print(" [MENU]   zurck")
        console_redraw()
        wait_click(0)
        if     is_pressed("set")     then call_data()
        elseif is_pressed("right") then
            local data = input_geo_data()
            if data ~= false then
                local id = table.getn(location) +1
                location[id] = {}
                location[id].name = data[1]
                location[id].lat  = data[2]
                location[id].lng  = data[3]
                location[id].utc  = data[4]
            end
        elseif is_pressed("left") then
            del_geo_data()
        elseif is_pressed("display") then
            cls()
            print("   GPS-Funktion nicht verfgbar!")
            console_redraw()
            sleep(2000)
        elseif is_pressed("menu")    then break
        end
    end
end

function input_geo_data()
    local idx = 1
    local geo = {"+", 0, 0, 0, 0, "+", 0, 0, 0, 0, 0}
    local pos = "\30"
    local res = false
    while true do
        local cur = {"_", "_", "_", "_", "_", "_", "_", "_", "_", "_", "_"}
        table.insert(cur, idx, pos)
        cls()
        print("     manuelle Eingabe Geo-Daten")
        print(line)
        print("     B", geo[1]..geo[2]..geo[3]..geo[4].."."..geo[5], " L", geo[6]..geo[7]..geo[8]..geo[9].."."..geo[10], "UTC", geo[11])
        print("      ", cur[1]..cur[2]..cur[3]..cur[4].." "..cur[5], "  ", cur[6]..cur[7]..cur[8]..cur[9].." "..cur[10], "   ", cur[11])
        print(line)
        print("   [\29]   Position  [\18] Wert")
        print("   [SET] Ok        [MENU] Abbruch")
        console_redraw()
        wait_click(0)
        if     is_pressed("right") then
            idx = idx + 1
            if idx > 11 then idx = 1 end
        elseif is_pressed("left") then
            idx = idx - 1
            if idx < 1 then idx = 11 end
        elseif is_pressed("up") then
            if idx == 1 or idx == 6 then
                if geo[idx] == "+" then geo[idx] = "-" else geo[idx] = "+" end
            elseif idx == 11 then
                geo[idx] = geo[idx] + 1
                if geo[idx] > 12 then geo[idx] = -12 end
            else
                geo[idx] = geo[idx] + 1
                if geo[idx] > 9 then geo[idx] = 0 end
            end
        elseif is_pressed("down") then
            if idx == 1 or idx == 6 then
                if geo[idx] == "+" then geo[idx] = "-" else geo[idx] = "+" end
            elseif idx == 11 then
                geo[idx] = geo[idx] - 1
                if geo[idx] < -12 then geo[idx] = 12 end
            else
                geo[idx] = geo[idx] - 1
                if geo[idx] < 0 then geo[idx] = 9 end
            end
        elseif is_pressed("set") then
            local lat = geo[2]*1000 + geo[3]*100 + geo[4]*10 + geo[5]
            if geo[1] == "-" then lat = lat * -1 end
            local lng = geo[7]*1000 + geo[8]*100 + geo[9]*10 + geo[10]
            if geo[6] == "-" then lng = lng * -1 end
            local utc = geo[11]
            res = {"Benutzerdaten", lat, lng, utc}
            break
        elseif is_pressed("menu") then
            res = false
            break
        end
    end
    return res
end

function del_geo_data()
    local id_max = table.getn(location)
    local id     = 1
    while true do
        cls()
        print("  Datensatz-Nr.:", id)
        print(" ", location[id].name)
        print(line)
        print("       [\18] Auswahl Datensatz")
        print("    [SET] Ok   [MENU] Abbruch")
        console_redraw()
        wait_click(0)
        if     is_pressed("up")   then
            id = id + 1
            if id > id_max then id = 1 end
        elseif is_pressed("down") then
            id = id - 1
            if id < 1 then id = id_max end
        elseif is_pressed("set") then
            if table.getn(location) > 1 then table.remove(location, id) end
            break
        elseif is_pressed("menu") then break
        end
    end
end

function is_leap_year(year)
    return year % 4 == 0 and (year % 100 ~= 0 or year % 400 == 0)
end

function get_day_of_year(year, month, day)
    local day_of_year = 0
    local feb = 28
    if is_leap_year(year) then feb = 29 end
    local days_in_month = {31, feb, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31}
    for i = 1, month-1 do
        day_of_year=day_of_year + days_in_month[i]
    end
    return day_of_year + day
end

function sun_times(doy, h, lat_deg, lng, utc)
    --based on http://lexikon.astronomie.info/zeitgleichung/
    --Deklination = 0.4095*sin(0.016906*(T-80.086))
    local D = cordic.integer_mul_div(cordic.sinr(cordic.integer_mul_div(16906, doy*1000-80086,1000000)), 4095, 10000)
    --Zeitgleichung: WOZ - MOZ = -0.171*sin(0.0337 * T + 0.465) - 0.1299*sin(0.01787 * T - 0.168)
    local WOZ = mul_1E3(cordic.sinr(((337 * doy) + (465*10))/10), -171)
    local MOZ = mul_1E3(cordic.sinr(((1787 * doy) - (168*100))/100), 130)
    local time_equation = WOZ - MOZ
    --Zeitdifferenz = 12*arccos((sin(h) - sin(B)*sin(Deklination)) / (cos(B)*cos(Deklination)))/Pi
    local time_diff = 12 * SCALE * cordic.acosr(div_1E3((cordic.sind(h) - mul_1E3(cordic.sind(lat_deg), cordic.sinr(D))), mul_1E3(cordic.cosd(lat_deg), cordic.cosr(D)))) / cordic.PI
    --Hchststand = 12 + Zeitzone - geographische Lnge /15 - Zeitgleichung
    local top = (12 + utc) * SCALE - div_1E3(lng * 100, 15 * SCALE) - time_equation
    local rising  = top - time_diff
    local setting = top + time_diff
    return rising, setting, top
end

function restore()
    cls()
    set_console_layout(0,0,25,5)
    set_console_autoredraw(1)
end

--MAIN-----------------------
cordic = require("cordicLib")
SCALE = cordic.SCALE

--date
year  = 2012 + y
if a ==1 then year  = os.date("%Y") end
month = m + 1
if a ==1 then month = os.date("%m") end
day   = d
if month == 2 and day > 28 then
    if is_leap_year(year) then
        day = 29
    else
        day = 28
    end
end
if a ==1 then day   = os.date("%d") end

--data
h1=-833  --sunset/sunrise h=-50'
h2=-6000 --civil twilight h= -6

location = {}
location[1] = {}
location[1].name = "Berlin" -- location
location[1].lat  = 525      -- latitude 52.5
location[1].lng  = 135      -- longitude 13.5
location[1].utc  = 1        -- time zone +1 h

--user interface
cls()
set_console_layout(5,3,40,14)
set_console_autoredraw(0)

line = "-----------------------------------" 
location_id = 1
DoY=get_day_of_year(year, month, day)

while true do
    location_id_max = table.getn(location)
    Utc = location[location_id].utc + s
    Utc_str = "UTC +"
    if Utc == 0 then Utc_str = "UTC +/-" end
    if Utc < 0 then Utc_str = "UTC " end
    Lat = location[location_id].lat
    Lat_str = string.format("%d.%d", Lat/10, math.abs(Lat%10))
    Lat_deg = Lat * 100
    Lng = location[location_id].lng
    Lng_str = string.format("%d.%d", Lng/10, math.abs(Lng%10))
    sunrise, sunset, suntop = sun_times(DoY, h1, Lat_deg, Lng, Utc)
    dawn, dusk = sun_times(DoY, h2, Lat_deg, Lng, Utc)
    cls()
    print(string.format(" [\18] %s %s%d  (%d)", location[location_id].name, Utc_str, location[location_id].utc, location_id))
    print(string.format("     %s %s %02d.%02d.%4d", Lat_str, Lng_str, day, month, year))
    print(line)
    print(string.format(" brgl. Morgendmmerung: %02d:%02d Uhr",dawn/1000, dawn%1000*600/100/100))
    print(string.format("          Sonnenaufgang: %02d:%02d Uhr",sunrise/1000, sunrise%1000*600/100/100))
    print(string.format("      Sonnenhchststand: %02d:%02d Uhr",suntop/1000, suntop%1000*600/100/100))
    print(string.format("        Sonnenuntergang: %02d:%02d Uhr",sunset/1000, sunset%1000*600/100/100))
    print(string.format(" brgl.  Abenddmmerung: %02d:%02d Uhr",dusk/1000, dusk%1000*600/100/100))
    print(line)
    print(" [SET] Geo-Daten   [MENU] beenden")
    console_redraw()
    wait_click(0)
    if     is_pressed("down") then
        location_id = location_id + 1
        if location_id > location_id_max then location_id = 1 end
    elseif is_pressed("up") then
        location_id = location_id - 1
        if location_id < 1 then location_id = location_id_max end
    elseif is_pressed("set") then
        menu_geo_data()
        location_id = 1
    elseif is_pressed("menu") then break
    end
end

restore()
